/****************************************************************************\
*                                                                            *
*  Storm.cpp                                                                 *
*  StormU.cpp                                                                *
*                                                                            *
*  Lightning storm demo by Diana Gruber. Illustrates Fastgraph's alpha       *
*  blending and gamma correction functionality.                              *
*                                                                            *
\****************************************************************************/

#include <vcl\vcl.h>
#pragma hdrstop

#include <math.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <time.h>

#include "StormU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TForm1::CreateParams(TCreateParams &Params)
{
   // add the CS_OWNDC style flag to the window class
   TForm::CreateParams(Params);
   Params.WindowClass.style |= CS_OWNDC;
}
//---------------------------------------------------------------------------

void __fastcall TForm1::FormActivate(TObject *Sender)
{
   fg_realize(hPal);
   Invalidate();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
   int bpp;
   time_t t;

   // make sure we can change the desktop resolution to 800x600
   Visible = True;
   bpp = fg_colors();
   if (fg_modetest(800,600,bpp) != 0)
   {
      MessageDlg("Cannot set 800x600 desktop.",
         mtError,TMsgDlgButtons()<<mbOK,0);
      Application->Terminate();
      return;
   }

   // make sure we're not using a 256-color or 16-color display
   if (bpp <= 8)
   {
      MessageDlg("High color or true color desktop required.",
         mtError,TMsgDlgButtons()<<mbOK,0);
      Application->Terminate();
      return;
   }
   fg_modeset(800,600,bpp,1);
   WindowState = wsMaximized;

   // initialize the device context and virtual palette
   hDC = GetDC(Form1->Handle);
   fg_setdc(hDC);
   hPal = fg_defpal();
   fg_realize(hPal);

   // initialize virtual buffers - use same color depth as display
   fg_vbinit();
   fg_vbdepth(bpp);

   vbHandle = fg_vballoc(1480,600);
   fg_vbopen(vbHandle);
   fg_vbcolors();

   vbExtra = fg_vballoc(800,600);
   fg_vbopen(vbExtra);
   fg_vbcolors();

   srand((unsigned int)time(&t));
   MakeOpacityBitmap();
   Flash1 = new Flash;
   Flash2 = new Flash;
   SoundOn = True;

   Intro();
   fg_mouseini();
   fg_mousevis(0);

   Application->OnActivate = OnActivate;

   // set animation timer
   m_timerID1 = timeSetEvent(40,10,AnimProc,0,TIME_PERIODIC);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormDestroy(TObject *Sender)
{
   if (m_timerID1) timeKillEvent(m_timerID1);
   PlaySound(NULL,NULL,0);
   fg_mousevis(1);
   fg_vbclose();
   fg_vbfree(vbHandle);
   fg_vbfree(vbExtra);
   fg_vbfin();
   fg_modeset(0,0,0,0);
   DeleteObject(hPal);
   ReleaseDC(Form1->Handle,hDC);
   Application->Minimize();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormKeyDown(TObject *Sender, WORD &Key,
   TShiftState Shift)
{
   if (Key == VK_ESCAPE)
   {
      Close();
   }
   else if (Key == 's' || Key == 'S')
   {
      SoundOn = !SoundOn;
      if (!SoundOn) PlaySound(NULL,NULL,0);
   }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormPaint(TObject *Sender)
{
   fg_vbopen(vbHandle);
   fg_vbpaste(340,1139,0,599,0,599);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Intro(void)
{
   // load the background graphic into the virtual buffers
   fg_vbopen(vbHandle);
   fg_move(340,0);
   fg_showjpeg("Background.jpg",FG_AT_XY);
   fg_vbpaste(340,1139,0,599,0,599);
   fg_vbcopy(340,1139,0,599,0,599,vbHandle,vbExtra);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::MakeOpacityBitmap()
{
   int i, x, y;
   int OpacityValue;
   int xDist, yDist;
   double Factor;

   // what we are creating here is an oval pattern to be used later with the
   // gamma adjustments to make the clouds glow around the lightning bolt
   i = 0;
   Factor = 3.0;
   for (y = 0; y < GlowHeight; y++)
   {
      yDist = abs(GlowHeight/2 - y);
      for (x = 0; x < GlowWidth; x++)
      {
         xDist = abs(GlowWidth/2 - x);
         OpacityValue = int(sqrt(Factor * (xDist*xDist + yDist*yDist*2)));
         Opacity[i++] = BYTE(255 - min(255,OpacityValue));
      }
   }
}
//---------------------------------------------------------------------------
static void CALLBACK AnimProc(UINT TimerID,UINT Msg,DWORD dwUser,DWORD dw1, DWORD dw2)
{
   // this is the callback function triggered by the animation timer
   if (TimerID == m_timerID1)
   {
      // repair the screen from the last frame
      fg_vbcopy(Form1->Flash1->Left-340,Form1->Flash1->Right-340,Form1->Flash1->Top,
      Form1->Flash1->Bottom,Form1->Flash1->Left,Form1->Flash1->Bottom,vbExtra,vbHandle);

      fg_vbcopy(Form1->Flash2->Left-340,Form1->Flash2->Right-340,Form1->Flash2->Top,
      Form1->Flash2->Bottom,Form1->Flash2->Left,Form1->Flash2->Bottom,vbExtra,vbHandle);

      // draw current flashes according to the current state
      Form1->Flash1->Action(Form1->Flash1);
      Form1->Flash2->Action(Form1->Flash2);

      // update the screen
      fg_vbpaste(Form1->Flash1->Left,Form1->Flash1->Right,Form1->Flash1->Top,
      Form1->Flash1->Bottom,Form1->Flash1->Left-340,Form1->Flash1->Bottom);

      fg_vbpaste(Form1->Flash2->Left,Form1->Flash2->Right,Form1->Flash2->Top,
      Form1->Flash2->Bottom,Form1->Flash2->Left-340,Form1->Flash2->Bottom);
   }
}
//---------------------------------------------------------------------------
Flash::Flash(void)
{
   // constructor
   CountDown = IRAND(9,18);
   Action = Waiting;
   Left = 0;
   Right = 0;
   Top = 0;
   Bottom = 0;
   BoltLeft = 0;
   BoltRight = 0;
   BoltTop = 0;
   BoltBottom = 0;
   BoltWidth = 0;
   BoltHeight = 0;
}
//---------------------------------------------------------------------------
void Flash::CalcAction(Node *node)
{
   int i;

   // when drawing the bolt: continue, branch, or stop
   i = IRAND(0,10);
   if (i < 2)
      node->action = END;
   else if (i < 7)
      node->action = CONTINUE;
   else
      node->action = BRANCH;
}
//---------------------------------------------------------------------------
void Flash::CalcFlash(void)
{
   // where and how big is the next flash
   x1 = IRAND(200,1279-GlowWidth);
   x1 = IRAND(600,900);
   x2 = x1 + GlowWidth - 1;
   y2 = IRAND(GlowHeight,550);
   if (y2 < 350) y2 += 200;
   y1 = y2 - GlowHeight + 1;
   GammaValue = IRAND(30,95) * 0.01;
   Left = x1;
   Right = x2;
   Top = y1;
   Bottom = y2;
}
//---------------------------------------------------------------------------
void Flash::StartFlash(class Flash *flash)
{
   int i;

   // state of the flash: beginning
   if (Form1->SoundOn)
   {
      i = IRAND(1,2);
      switch (i)
      {
         case 1:
            PlaySound("Thunder1.wav",NULL,SND_FILENAME|SND_NOSTOP|SND_ASYNC);
            break;
         case 2:
            PlaySound("Thunder2.wav",NULL,SND_FILENAME|SND_NOSTOP|SND_ASYNC);
            break;
      }
   }

   flash->CalcFlash();
   flash->ShowFlash();

   flash->CalcAction(&root);
   root.x = flash->x1 + GlowWidth/2;
   root.y = flash->y1 + GlowHeight/2 -50;
   flash->BoltLeft = root.x;
   flash->BoltRight = root.x;
   flash->BoltTop = root.y;
   flash->BoltBottom = root.y;
   root.distance = 0;
   flash->ShowBolt(&root);
   StopAll = False;

   flash->BoltHeight = flash->BoltBottom - flash->BoltTop + 1;
   flash->BoltWidth = flash->BoltRight - flash->BoltLeft + 1;
   if (flash->BoltHeight > 1)
   {
      fg_setcolor(fg_maprgb(255,255,230));
      fg_move(flash->BoltLeft,flash->BoltBottom);
      fg_getmap(flash->Bolt,flash->BoltWidth,flash->BoltHeight);
   }

   flash->CountDown = 4;

   // increment the state
   flash->Action = Flashing0;
}
//---------------------------------------------------------------------------
void Flash::Flashing0(class Flash *flash)
{
   // state of the flash: next state
   flash->GammaValue *= 0.7;
   flash->ShowFlash();

   if (flash->BoltHeight > 1)
   {
      fg_setcolor(fg_maprgb(250,250,220));
      fg_move(flash->BoltLeft,flash->BoltBottom);
      fg_drawmap(flash->Bolt,flash->BoltWidth,flash->BoltHeight);
   }

   // increment the state
   flash->Action = Flashing1;
}
//---------------------------------------------------------------------------
void Flash::Flashing1(class Flash *flash)
{
   flash->GammaValue *= 0.7;
   flash->ShowFlash();
   if (flash->BoltHeight > 1)
   {
      fg_setcolor(fg_maprgb(210,210,210));
      fg_move(flash->BoltLeft,flash->BoltBottom);
      fg_drawmap(flash->Bolt,flash->BoltWidth,flash->BoltHeight);
   }

   // increment the state
   flash->Action = Flashing2;
}
//---------------------------------------------------------------------------
void Flash::Flashing2(class Flash *flash)
{
   flash->ShowFlash();
   if (flash->BoltHeight > 1)
   {
      fg_setcolor(fg_maprgb(175,175,190));
      fg_move(flash->BoltLeft,flash->BoltBottom);
      fg_drawmap(flash->Bolt,flash->BoltWidth,flash->BoltHeight);
   }

   // increment the state
   flash->Action = Flashing3;
}
//---------------------------------------------------------------------------
void Flash::Flashing3(class Flash *flash)
{
   // this should clear the flash
   flash->Action = EndFlash;
}
//---------------------------------------------------------------------------
void Flash::EndFlash(class Flash *flash)
{
   // wait a while before starting the next flash
   flash->CountDown = IRAND(30,90);
   flash->Action = Waiting;
   flash->Left = 0;
   flash->Right = 0;
   flash->Top = 0;
   flash->Bottom = 0;
   flash->BoltLeft = 0;
   flash->BoltRight = 0;
   flash->BoltTop = 0;
   flash->BoltBottom = 0;
   flash->BoltWidth = 0;
   flash->BoltHeight = 0;
}
//---------------------------------------------------------------------------
void Flash::ShowFlash(void)
{
   // grab the area under the flash, do a gamma correction with an
   // oval pattern, blend it with the background. This gives you
   // an incremental glow... start out bright, then fade.
   fg_move(x1,y2);
   fg_getdcb(Foreground,GlowWidth,GlowHeight);
   fg_gammadcb(Foreground,Foreground,GammaValue,GlowWidth*GlowHeight);
   fg_blendvbv(Foreground,NULL,Form1->Opacity,GlowWidth,GlowHeight);
}
//---------------------------------------------------------------------------
void Flash::ShowBolt(Node *node)
{
   Node NewNode;
   int Factor;
   int xDist, yDist, nDist;

   // Recursive function. Each node has 3 states: continue, end or branch.
   // If the current state is continue or branch, this function will be
   // called recursively.

   if (StopAll)
      return;
   if (node->distance > 11)
      node->action = END;

   Factor = (node->distance + 3) / 3;
   fg_setcolor(fg_maprgb(255,255,230));
   switch (node->action)
   {
      case END:
         return;

      case CONTINUE:
         NewNode.distance = node->distance;
         CalcAction(&NewNode);
         NewNode.x = node->x + (IRAND(0,20)-10) / Factor;
         NewNode.y = node->y + (IRAND (5,20)) / Factor;
         if (NewNode.y > 500)
         {
            NewNode.action = END;
            StopAll = True;
         }
         fg_move(node->x,node->y);
         fg_draw(NewNode.x,NewNode.y);

         Bottom = max(NewNode.y,Bottom);
         BoltBottom = max(NewNode.y,BoltBottom);
         Left = min(NewNode.x,Left);
         BoltLeft = min(NewNode.x,BoltLeft);
         Right = max(NewNode.x,Right);
         BoltRight = max(NewNode.x,BoltRight);

         ShowBolt(&NewNode);
         break;

      case BRANCH:
         // calculate properties for first new node
         CalcAction(&NewNode);
         nDist = IRAND(0,3);
         NewNode.distance = node->distance+nDist;
         xDist = IRAND(5,15);
         NewNode.x = node->x - xDist/Factor;
         yDist = IRAND(10,30);
         NewNode.y = node->y + yDist/Factor;

         Bottom = max(NewNode.y,Bottom);
         BoltBottom = max(NewNode.y,BoltBottom);
         Left = min(NewNode.x,Left);
         BoltLeft = min(NewNode.x,BoltLeft);
         Right = max(NewNode.x,Right);
         BoltRight = max(NewNode.x,BoltRight);

         if (NewNode.y > 500)
         {
            NewNode.action = END;
            StopAll = True;
         }

         // draw the new node
         fg_move(node->x,node->y);
         fg_draw(NewNode.x,NewNode.y);

         // recurse this side
         ShowBolt(&NewNode);

         // calculate properties for new node on the other side
         CalcAction(&NewNode);
         NewNode.distance = node->distance + (3-nDist);
         NewNode.x = node->x + (20-xDist) / Factor;
         NewNode.y = node->y + (60-yDist) / Factor;

         Bottom = max(NewNode.y,Bottom);
         BoltBottom = max(NewNode.y,BoltBottom);
         Left = min(NewNode.x,Left);
         BoltLeft = min(NewNode.x,BoltLeft);
         Right = max(NewNode.x,Right);
         BoltRight = max(NewNode.x,BoltRight);

         if (NewNode.y > 500)
         {
            NewNode.action = END;
            StopAll = True;
         }

         // draw the node on the other side
         fg_move(node->x,node->y);
         fg_draw(NewNode.x,NewNode.y);

         // recurse the other side
         ShowBolt(&NewNode);
         break;
   }
}
//---------------------------------------------------------------------------
void Flash::Waiting(class Flash *flash)
{
   // no current flash. wait a while to start a new one.
   flash->CountDown--;
   if (flash->CountDown <= 0) flash->Action = StartFlash;
}
